import numpy as np
from cmdstanpy import CmdStanModel
from scipy.io import loadmat
from saveEstimates import save_estimates


path2data = '/path/to/data/' # path to data (observations and sea-level fingerprint)
path2code = 'path/to/code/' # path to Stan code (bhm_sea_level.stan)
pathOut = '/path/to/output/' # path where the output of the Stan model will be saved
modelName = 'bhm_model.stan' # name of the file containing the Stan model
inFile = 'data_input.mat' # name of the file containing the input data (available at https://doi.org/10.5281/zenodo.6797928)
outFile = 'estimates_bhm.mat' # name of the file where the output from the Stan model will be saved (mat format)
K = 5 # number of retained EOFs
M = 20 # number of neighbors for the NNGPs

data = loadmat(path2data + inFile)
y = data["tg_sl"] # T x N matrix containing the annual sea levels from tide gauges (T: temporal samples; N: stations)
lon = data["lon_tg"][:,0] # N x 1 vector containing the longitudes of the tide gauge sites
lat = data["lat_tg"][:,0] # N x 1 vector containing the latitudes of the tide gauge sites
eof = data["eof_tg"][:,0:K] # N x K matrix containing the EOFs at tide gauge sites (K: number of retained EOFs)
ya = data["sla"] # Nx x Ny x Ta matrix containing the gridded annual sea levels from altimetry (Nx: longitude; Ny: latitude; Ta: temporal samples)
lona = data["lon"] # Nx x Ny matrix containing the longitudes of the altimetry grid points
lata = data["lat"] # Nx x Ny matrix containing the latitudes of the altimetry grid points
eofa = data["eof"][:,:,0:K] # Nx x Ny x K matrix containing the EOFs on the altimetry grid
mur = data["grd_mean_rsl_tg"] # N x T matrix containing the ensemble-mean rates of RSL change due to GRD effects at tide gauge sites
mu_giar = data["gia_mean_rsl_tg"][:,0] # N x 1 vector containing the ensemble-mean rates of RSL change due to GIA at tide gauge sites
Lr = data["grd_chol_rsl_tg"] # N x N matrix containing the cholesky factor (lower triangular) for the covariance matrix (ensemble of RSL fingerprints)
Lr0 = data["grd_chol_rsl_init_tg"] # Same as Lr but for the initial state of the spatiotemporal random walk
Lgiar = data["gia_chol_rsl_tg"] # N x N matrix containing the cholesky factor for the covariance matrix (ensemble of RSL rates due to GIA)
mug = data["grd_mean_gsl"][:,:,-ya.shape[2]:] # Nx x Ny x Ta matrix containing the ensemble-mean rates of GSL change due to GRD effects on the altimetry grid
mu_giag = data["gia_mean_gsl"] # Nx x Ny matrix containing the ensemble-mean rates of GSL change due to GIA on the altimetry grid
Lg = data["grd_chol_gsl"] # Na x Na matrix containing the cholesky factor for the covariance matrix (ensemble of GSL fingerprints) (Na: grid points with valid altimetry data)
Lg0 = data["grd_chol_gsl_init"] # Same as Lg0 but for the initial state of the spatiotemporal random walk
Lgiag = data["gia_chol_gsl"] # Na x Na matrix containing the cholesky factor for the covariance matrix (ensemble of GSL rates due to GIA)

# extract all the GSL fields at grid points with valid altimetry data
n1,n2,n3 = ya.shape
ya = np.reshape(ya,(n1*n2,n3),order='F')
eofa = np.reshape(eofa,(n1*n2,eofa.shape[2]),order='F')
k = np.nonzero(np.mean(ya,1) == np.mean(ya,1))[0] # indices of grid points with valid altimetry data
ya = ya[k,:]
ya = ya.T
eofa = eofa[k,:]
lona = (lona.flatten('F'))[k]
lata = (lata.flatten('F'))[k]
mug = np.reshape(mug,(n1*n2,n3),order='F')
mug = mug[k,:]
mu_giag = mu_giag.flatten('F')[k]

# flatten matrices containing observational data for input to Stan
nt,n = y.shape
nta,na = ya.shape
y_flat = y.flatten('C') # row-major order (matrices in Stan store data in column-major order)
kok = np.nonzero(y_flat == y_flat)[0]
y_ok = y_flat[kok]
nok = y_ok.shape[0]
kok = kok + 1 # Python uses zero-based indexing while in Stan indices are one-based
ya_flat = ya.flatten('C') 
koka = np.nonzero(ya_flat == ya_flat)[0]
ya_ok = ya_flat[koka]
noka = ya_ok.shape[0]
koka = koka + 1

# store all input data into dictionary
modelData = {'ntg':n,'na':na,'nt':nt,'nta':nta,'nok':nok,'noka':noka,'M':M,'K':K,'lon_tg':lon,'lat_tg':lat,'lona':lona,'lata':lata,'kok':kok,'koka':koka,'y_tg':y_ok,\
             'y_alti':ya_ok,'eof':eof,'eofa':eofa,'mur':mur,'mug':mug,'mu_giar':mu_giar,'mu_giag':mu_giag,'Lr':Lr,'Lg':Lg,'Lr0':Lr0,'Lg0':Lg0,'Lgiar':Lgiar,'Lgiag':Lgiag,'ya':ya}
             
bhm_model = CmdStanModel(stan_file=path2code + modelName,cpp_options={'STAN_THREADS':False}) # compile model
fit = bhm_model.sample(data=modelData.copy(),chains=4,iter_sampling=750,iter_warmup=500,show_progress=True,inits=0.5,max_treedepth=11,adapt_delta=0.8,output_dir = pathOut) # fit BHM 

# The variable 'fit' is a stanfit object that contains the output derived from fitting the max-stable model to the data 
# The following function call extracts the data in 'fit' and saves them to a file:
save_estimates(fit,modelData,pathOut,outFile) # Save output of the BHM to file (mat format)

# Once the solutions from the Stan model are available, predictions on a finer grid can be made by running the Python script: bhm_prediction.py

